# $Id: __init__.py,v 1.1.1.1 1999/10/27 10:21:39 zen Exp $
#
__doc__ = '''
$Id: __init__.py,v 1.1.1.1 1999/10/27 10:21:39 zen Exp $
'''

import GenericUserFolder

def initialize(context):
    try:
	context.registerClass(
	    GenericUserFolder.GenericUserFolder,
	    constructors=(GenericUserFolder.manage_addGenericUserFolderForm,
			  GenericUserFolder.manage_addGenericUserFolder),
	    icon='folder.gif'
	    )
	context.registerBaseClass(GenericUserFolder.GenericUserFolder)
	context.registerBaseClass(GenericUserFolder.User)

    except:
        """If you can't register the product, tell someone.

        Zope will sometimes provide you with access to "broken product" and
        a backtrace of what went wrong, but not always; I think that only
        works for errors caught in your main product module.

        This code provides traceback for anything that happened in
        registerClass(), assuming you're running Zope in debug mode."""

        import sys, traceback, string
        type, val, tb = sys.exc_info()
        sys.stderr.write(string.join(
	    traceback.format_exception(type, val, tb),''))
        del type, val, tb

